
//#define _BSD_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/ip_icmp.h>
#include <netdb.h>
#include <sys/time.h>
#include <signal.h>
#include <unistd.h>

#define BUFSIZE 3000

/* global variables */
int sd;                      /* socket descriptor */
pid_t pid;                   /* PID */
struct sockaddr_in servaddr; /* sockaddr() for sending   */
struct sockaddr_in from;     /* sockaddr() for receiving */

double tmin = 999999999.0;   /* minimal round-trip time */
double tmax = 0;             /* maximum round-trip time */
double tsum = 0;             /* summary time for calculating average time */
int nsent = 0;               /* number of sended packages */
int nreceived = 0;           /* number of received packages */

/* prototypes */
void pinger(void);
void output(char *, int, struct timeval *);
void catcher(int);
void tv_sub(struct timeval *, struct timeval *);
unsigned short in_cksum(unsigned short *, int);

/* check control sum */
unsigned short in_cksum(unsigned short *addr, int len)
{
	unsigned short result;
	unsigned int sum = 0;
	
	/* sum for all dwords */
	while (len > 1) {
		sum += *addr++;
		len -= 2;
	}
	
	/* add byte to sum */
	if (len == 1) {
		sum += *(unsigned char*) addr;
	}
	
	sum = (sum >> 16) + (sum & 0xFFFF); /* carrying */
	sum += (sum >> 16);                 /* again */
	result = ~sum;                      /* and invert */
	return result;
}

/* timeval sub routine */
void tv_sub(struct timeval *out, struct timeval *in)
{
	if ( (out->tv_usec -= in->tv_usec) < 0) {
		out->tv_sec--;
		out->tv_usec += 1000000;
	}
	out->tv_sec -= in->tv_sec;
}

/* package parsing and output */
void output(char *ptr, int len, struct timeval *tvrecv)
{
	int iplen;
	int icmplen;
	struct ip *ip;
	struct icmp *icmp;
	struct timeval *tvsend;
	double rtt;
	
	ip = (struct ip *) ptr; /* beginning of IP header */
	iplen = ip->ip_hl << 2; /* length of IP header */
	
	icmp = (struct icmp *) (ptr + iplen); /* beginning of ICMP header */
	if ( (icmplen = len - iplen) < 8) {   /* length of ICMP header */
		fprintf(stderr, "icmplen (%d) < 8", icmplen);
	}
	
	if (icmp->icmp_type == ICMP_ECHOREPLY) {
		if (icmp->icmp_id != pid) {
			/* this is ECHO REPLY for another process */
			return;
		}
		
		tvsend = (struct timeval *) icmp->icmp_data;
		tv_sub(tvrecv, tvsend);
		
		/* round-trip time */
		rtt = tvrecv->tv_sec * 1000.0 + tvrecv->tv_usec / 1000.0;
		
		nreceived++;
		
		tsum += rtt;
		if (rtt < tmin) {
			tmin = rtt;
		}
		if (rtt > tmax) {
			tmax = rtt;
		}
		
		printf("%d bytes from %s: icmp_seq=%u, ttl=%d, time=%.3f ms\n", icmplen, inet_ntoa(from.sin_addr), icmp->icmp_seq, ip->ip_ttl, rtt);
	}
}

/* package preparing and sending ICMP ECHO REQUEST */
void pinger(void)
{
	int icmplen;
	struct icmp *icmp;
	char sendbuf[BUFSIZE];
	
	icmp = (struct icmp *) sendbuf;
	
	/* prepare ICMP fields for package */
	icmp->icmp_type = ICMP_ECHO;
	icmp->icmp_code = 0;
	icmp->icmp_id = pid;
	icmp->icmp_seq = nsent++;
	gettimeofday((struct timeval *) icmp->icmp_data, NULL);
	
	/* length of ICMP header: 8(control) + 56(data) */
	icmplen = 8 + 56;
	/* check sum of ICMP header and data */
	icmp->icmp_cksum = 0;
	icmp->icmp_cksum = in_cksum((unsigned short *) icmp, icmplen);
	
	if (sendto(sd, sendbuf, icmplen, 0, (struct sockaddr *)&servaddr, sizeof(servaddr)) < 0) {
		perror("sendto() failed");
		exit(-1);
	}
}

/* signal handler for SIGALRM and SIGINT */
void catcher(int signum)
{
	if (signum == SIGALRM)
	{
		pinger();
		return;
	} else if (signum == SIGINT) {
		printf("\n--- %s ping statistics ---\n", inet_ntoa(servaddr.sin_addr));
		printf("%d packets transmitted, ", nsent);
		printf("%d packets received, ", nreceived);
		if (nsent) {
			if (nreceived > nsent) {
				printf("-- somebody's printing up packets!");
			} else {
				printf("%d%% packet loss", (int) (((nsent-nreceived)*100)/nsent));
			}
		}
		printf("\n");
		if (nreceived) {
			printf("round-trip min/avg/max = %.3f/%.3f/%.3f ms\n", tmin, tsum / nreceived, tmax);
		}
		fflush(stdout);
		exit(-1);
	}
}

/* entry point */
int main(int argc, char *argv[])
{
	int size;
	socklen_t fromlen;
	int n;
	struct timeval tval;
	char recvbuf[BUFSIZE];
	struct hostent *hp;
	struct sigaction act;
	struct itimerval timer;
	const int on = 1;
	
	if (argc != 2) {
		fprintf(stderr, "Usage: %s <hostname>\n", argv[0]);
		exit(-1);
	}
	
	pid = getpid();
	
	/* signal handler setup */
	memset(&act, 0, sizeof(act));
	/* setup catcher() for handler */
	act.sa_handler = &catcher;
	sigaction(SIGALRM, &act, NULL);
	sigaction(SIGINT, &act, NULL);
	
	if ((hp = gethostbyname(argv[1])) == NULL) {
		herror("gethostbyname() failed");
		exit(-1);
	}
	
	if ((sd = socket(PF_INET, SOCK_RAW, IPPROTO_ICMP)) < 0) {
		perror("socket() failed");
		exit(-1);
	}
	
	/* recovery permissions */
	setuid(getuid());
	
	/* enable broadcast messages support */
	setsockopt(sd, SOL_SOCKET, SO_BROADCAST, &on, sizeof(on));
	
	/* setup size of receiving buffer */
	size = 60 *1024;
	setsockopt(sd, SOL_SOCKET, SO_RCVBUF, &size, sizeof(size));
	
	/* run SIGALRM timer after 1 usec */
	timer.it_value.tv_sec=0;
	timer.it_value.tv_usec=1;
	/* repeat timer each 1 sec */
	timer.it_interval.tv_sec=1;
	timer.it_interval.tv_usec=0;
	/* run real time timer */
	setitimer(ITIMER_REAL, &timer, NULL);
	
	bzero(&servaddr, sizeof(servaddr));
	servaddr.sin_family = AF_INET;
	servaddr.sin_addr= *((struct in_addr *) hp->h_addr);
	
	fromlen = sizeof(from);
	
	/* infinite loop for receiving packages */
	while (1) {
		n = recvfrom(sd, recvbuf, sizeof(recvbuf), 0, (struct sockaddr *)&from, &fromlen);
		
		if (n < 0) {
			if (errno == EINTR) {
				continue;
			}
			perror("recvfrom() failed");
			continue;
		}
		
		/* get current system time */
		gettimeofday(&tval, NULL);
		
		/* callback function for parcing output */
		output(recvbuf, n, &tval);
	}
	
	return 0;
}

